
let compose f g x = (f (g x))

let rec ppTerm t = match t with
  Var x -> x |
  Fn (name,[]) -> name |
  Fn (name,args) -> name ^ "(" ^ (String.concat "," (map ppTerm args)) ^ ")"


let ppAtom a = match a with
  R (name,[]) -> name |
  R (name,args) -> name ^ "(" ^ (String.concat "," (map ppTerm args)) ^ ")"

let rec ppFormula f = match f with
 Not n -> "~("^(ppFormula n)^")" |
 Atom a -> ppAtom a |
 Forall (var, sf) -> "ALL " ^ var ^ ". (" ^ (ppFormula sf) ^ ")" |
 Exists (var, sf) -> "EX " ^ var ^ ". (" ^ (ppFormula sf) ^ ")" |
 Iff (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") <=> (" ^ (ppFormula rhs) ^ ")" |
 Imp (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") => (" ^ (ppFormula rhs) ^ ")" |
 And (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") /\ (" ^ (ppFormula rhs) ^ ")" |
 Or (lhs,rhs) -> "(" ^ (ppFormula lhs) ^ ") \/ (" ^ (ppFormula rhs) ^ ")" |
 True -> "True" |
 False -> "False"

let ppTLL tll = "[" ^ (String.concat ", " (map (fun ss -> "[" ^ (String.concat "," (map ppFormula ss)) ^ "]") tll)) ^ "]"


let rec escape s = if String.contains s '_'
 then (String.set s (String.index s '_') '-' ; escape s)
 else s

let norm f = simpcnf (skolemize (pnf (Not(generalize f)))) ;;

let ourRes cnfl = []

let rec formatTerm t = match t with
        Var x -> x^":S" |
        Fn (name,args) -> match args with
                              [] -> (escape name) |
                              _  -> (escape name)
                                   ^"("^(String.concat "," (map formatTerm args))^")"

let rec formatAtom a = match a with
        Not  x -> formatAtom x |
        Atom (R (name,args)) -> (match args with
                                  [] -> escape name |
                                  _  -> (escape name)^"(" ^(String.concat "," (map formatTerm args)) ^")") |
        Forall _ -> failwith "Unhandled Forall" |
   _ -> failwith "Unhandled Case: formatAtom"

let isNot a = match a with Not _ -> true | _ -> false

let fixEmptyMaps s = match s with
  "" -> "(empty).AtomMagma" |
  _ -> s

let formatClause cl = "(" ^ (fixEmptyMaps (String.concat "," (map formatAtom (filter isNot cl))))
    ^" -> "^
   (fixEmptyMaps (String.concat "," (map formatAtom (filter (compose not isNot) cl)))) ^ ")"

let formatFormula f = "red ?((\n  " ^ (String.concat ",\n  " (map formatClause f)) ^ "\n)) .\n\nq .\n";;

let nonCrapEq (n1,l1) (n2,l2) = ((String.compare n1 n2) == 0) && (l1 = l2);;

let rec uniq l = match l with
 [] -> [] |
 (x::xs) -> (x::uniq (filter (fun y -> not (nonCrapEq x y)) xs));;

let replicate n x = match n with
 0 -> [] |
 _ -> (x::replicate (n-1) x)

let rec getPred a = match a with
       Atom (R (name,args)) -> (escape name,length args) |
       Not a -> getPred a |
       _ -> failwith "Unhandled Case: getPred"

let predNumber = ref 0

let formatPreds f = let preds = uniq (List.concat (map (map getPred) f)) in
 String.concat "\n" (map (fun (n,l) -> "  op " ^ n ^ " : " ^
 (String.concat " " (replicate l "S")) ^ " -> Atom [metadata \""^ 
 (string_of_int (incr predNumber; !predNumber))
 ^"\"] .") preds)

let rec getFuncs t = match t with
 Var _ -> [] |
 Fn (name,args) -> ((escape name,length args) :: (List.concat (map getFuncs args)))

let rec getFuncsP a = match a with
 Not x -> getFuncsP x |
 Atom (R (_,args)) -> List.concat (map getFuncs args) |
 _ -> failwith "Unhandled Case: getFuncsP"

let funcNumber = ref 0

let formatFuncs f = let funcs = uniq (List.concat (List.concat (map (map (getFuncsP)) f))) in
 String.concat "\n" (map (fun (n,l) -> "  op " ^ n ^ " : " ^
 (String.concat " " (replicate l "S")) ^ " -> S [metadata \""^ 
 (string_of_int (incr funcNumber; !funcNumber))
 ^"\"] .") funcs)

let formatDecls f = (formatPreds f) ^ "\n" ^ (formatFuncs f)

let modname name = String.uppercase (escape name)

let usefulTime f x = 
  let start_time = Sys.time() in
  let result = f x in
  let finish_time = Sys.time() in
  (finish_time -. start_time,result);;

let formatMain name f = let cnf = norm f 
in 
"load ../../../rtp\n\nfmod " ^ (modname name) ^ 
" is\npr PRED-BASE .\n\nsort S .\n" ^ formatDecls cnf ^ "\nendfm\n\n" ^
"fmod TEST is\n  inc RTP .\n  inc " ^ (modname name) ^ " .\n\n" ^
"  eq USER-MODULE-NAME = \'" ^ (modname name) ^ " .\nendfm\n\n" ^
(formatFormula cnf);;

let trans name f = (try (let format = (formatMain name f) and chan = open_out (name ^".maude") in 
 (print_string "Success\n"; output_string chan format; flush chan; close_out chan))
with Failure e -> 
(let chan = open_out (name ^".maude") in (print_string "Failure\n"; output_string chan 
(let cnf = norm f in
("load ../../../rtp\n\nfmod " ^ (modname name) ^ 
" is\npr PRED-BASE .\n\nsort S .\n" ^ formatDecls cnf ^ "\nendfm\n\n" ^
"fmod TEST is\n  inc RTP .\n  inc " ^ (modname name) ^ " .\n\n" ^
"  eq USER-MODULE-NAME = \'" ^ (modname name) ^ " .\nendfm\n\n" ^
(formatFormula cnf))))));;
